#include "defs.h"

double betweennessCentrality_simple(graph* G, DOUBLE_T* BC) {

    LONG_T i, j, k, p, n, chunkSize;
    LONG_T tid, nthreads;
    omp_lock_t* vLock;
    VERT_T* S;          /* stack of vertices in the order of non-decreasing 
                           distance from s. Also used for implicitly representing 
                           the BFS queue */
    dynArray* P;        /* predecessors of a vertex v on shortest paths from s */
    LONG_T* sig;        /* No. of shortest paths */
    LONG_T* d;          /* Length of shortest path between every pair */
    DOUBLE_T* del;      /* dependency of vertices */
    LONG_T* permV; 
    LONG_T start, end;
    LONG_T v, w;
    LONG_T numV;
    double time;
    double sum;

/* The outer loop is parallelized in this case. Each thread does a BFS 
   and the BC values are incremented atomically */   
#pragma omp parallel \
private(tid, nthreads, S, P, sig, del, d, start, end, v, w, numV) \
private(i, j, k, p, n)
{
    tid = omp_get_thread_num();
    nthreads = omp_get_num_threads();
     
    /* numV: no. of vertices to run BFS from = 2^K4Approx */
    numV = 1<<K4Approx;
    n = G->n;

    if (tid == 0) {
        permV = (LONG_T *) malloc(n*sizeof(LONG_T)); 
        chunkSize = n/nthreads;
        vLock = (omp_lock_t *) malloc(n*sizeof(omp_lock_t));
    }
   
    #pragma omp barrier

    /* permute the vertices and use numV among them to run BFS from */
    #pragma omp for schedule(static, chunkSize)
    for (i=0; i<n; i++) {
        permV[i] = i;
    }

    if (tid == 0) {
        srand48(SCALE*387543);
        for (i=0; i<n; i++) {
            j = lrand48() % n;
            /* Swap permV[i] and permV[j] */
            k = permV[i];
            permV[i] = permV[j];
            permV[j] = k;
        } 
    }
    
    #pragma omp barrier
    
    if (tid == 0) {
        time = omp_get_wtime();
        chunkSize = n/nthreads;
    }

    #pragma omp barrier

    #pragma omp for schedule(static, chunkSize)
    for (i=0; i<n; i++) {
        omp_init_lock(&vLock[i]);
    }

    S   = (VERT_T *) malloc(n*sizeof(VERT_T));
    P   = (dynArray *) calloc(n, sizeof(dynArray));
    sig = (LONG_T *) malloc(n*sizeof(LONG_T));
    d   = (LONG_T *) malloc(n*sizeof(LONG_T));
    del = (DOUBLE_T *) calloc(n, sizeof(DOUBLE_T));
    
    for (i=0; i<n; i++) {
        initDynArray(&P[i]);    
        d[i] = -1;    
    }

    #pragma omp barrier
    
    #pragma omp for schedule(dynamic)
    for (p=0; p<numV; p++) {
        i = permV[p];
        sig[i] = 1;
        d[i] = 0;

        start = 0;
        S[0] = i;
        end = 1;

        while (end - start > 0) {

            v = S[start];

            for (j=G->numEdges[v]; j<G->numEdges[v+1]; j++) {

                /* Filter edges with weights divisible by 8 */
                if ((G->weight[j] & 7) != 0) {
                    w = G->endV[j];
                    if (v != w) {

                        /* w found for the first time? */ 
                        if (d[w] < 0) {
                            S[end++] = w;
                            d[w] = d[v] + 1;
                            sig[w] = 0;
                        }

                        if (d[w] == d[v] + 1) {
                            sig[w] = sig[w] + sig[v];
                            dynArrayInsert(&P[w], v);
                        }

                    }
                }
            }

            start++;
                
        }

        for (j=end-1; j>0; j--) {
            w = S[j];
            for (k = 0; k<P[w].count; k++) {
                v = P[w].vals[k];
                del[v] = del[v] + sig[v]*(1+del[w])/sig[w];
            }
            /* increment BC value atomically */
            omp_set_lock(&vLock[w]);
            BC[w] += del[w];
            omp_unset_lock(&vLock[w]);
        } 

        for (j=end-1; j>=0; j--) {
            w = S[j];
            d[w] = -1;
            del[w] = 0;
            P[w].count = 0;
        }
    }

    #pragma omp for schedule(static, chunkSize)
    for (i=0; i<n; i++) {
        omp_destroy_lock(&vLock[i]);
    }
    
    free(S);
    for (j=0; j<n; j++)
        freeDynArray(&P[j]);
    free(P);
    free(sig);
    free(d);
    free(del);
   
    if (tid == 0) { 
        free(vLock);
        time = omp_get_wtime() - time;
        free(permV);
    }

}    

    /* Verification */
    sum = 0;
    n = G->n;
    numV = 1<<K4Approx;
    for (i=0; i<n; i++) {
        sum += BC[i];
        // fprintf(stderr, "%lf ", BC[i]);
    }
    fprintf(stderr, "BC sum: %lf\n", sum); 
    
    return time;
}

double betweennessCentrality_parBFS(graph* G, DOUBLE_T* BC) {

    LONG_T i, j, k, p, n, chunkSize;
    LONG_T tid, nthreads;
    omp_lock_t* vLock;
    VERT_T *S, *myS;    /* stack of vertices in the order of non-decreasing 
                           distance from s. Also used for implicitly representing 
                           the BFS queue */
    dynArray* P;        /* predecessors of a vertex v on shortest paths from s */
    LONG_T* sig;        /* No. of shortest paths */
    LONG_T* d;          /* Length of shortest path between every pair */
    DOUBLE_T* del;      /* dependency of vertices */
    LONG_T* permV; 
    LONG_T *start, *end;
    LONG_T *psCount, myCount;
    LONG_T v, w, vert;
    int myLock;
    LONG_T numV, count, numPhases;
    LONG_T MAX_NUM_PHASES;
    double time;
    double sum;


#pragma omp parallel \
private(tid, nthreads, v, vert, w, numV, myS, myCount, myLock,numPhases) \
private(i, j, k, p, n, count)
{
    tid = omp_get_thread_num();
    nthreads = omp_get_num_threads();
     
    /* numV: no. of vertices to run BFS from = 2^K4Approx */
    numV = 1<<K4Approx;
    n = G->n;

    myCount = 0;
    MAX_NUM_PHASES = 100;
    
    
    if (tid == 0) {
        permV = (LONG_T *) malloc(n*sizeof(LONG_T)); 
        chunkSize = n/nthreads;
        vLock = (omp_lock_t *) malloc(n*sizeof(omp_lock_t));
    }
   
    #pragma omp barrier

    /* permute the vertices and use numV among them to run BFS from */
    #pragma omp for schedule(static, chunkSize)
    for (i=0; i<n; i++) {
        permV[i] = i;
    }

    if (tid == 0) {
        srand48(SCALE*387543);
        for (i=0; i<n; i++) {
            j = lrand48() % n;
            /* Swap permV[i] and permV[j] */
            k = permV[i];
            permV[i] = permV[j];
            permV[j] = k;
        } 

    }
    
    #pragma omp barrier
    
    if (tid == 0) {
        time = omp_get_wtime();
        chunkSize = n/nthreads;
    }

    #pragma omp barrier

    #pragma omp for schedule(static, chunkSize)
    for (i=0; i<n; i++) {
        omp_init_lock(&vLock[i]);
    }

    if (tid == 0) {
        S   = (VERT_T *) malloc(n*sizeof(VERT_T));
        P   = (dynArray *) calloc(n, sizeof(dynArray));
        sig = (LONG_T *) malloc(n*sizeof(LONG_T));
        d   = (LONG_T *) malloc(n*sizeof(LONG_T));
        del = (DOUBLE_T *) calloc(n, sizeof(DOUBLE_T));
        start = (LONG_T *) malloc(MAX_NUM_PHASES*sizeof(LONG_T));
        end = (LONG_T *) malloc(MAX_NUM_PHASES*sizeof(LONG_T));
        psCount = (LONG_T *) malloc((nthreads+1)*sizeof(LONG_T));
    }
   
    myS = (LONG_T *) malloc(n*sizeof(LONG_T));
    myCount = 0;
    
    #pragma omp barrier
     
    #pragma omp for schedule(static, chunkSize) 
    for (i=0; i<n; i++) {
        initDynArray(&P[i]);
        d[i] = -1;
    }

    #pragma omp barrier
    
    for (p=0; p<numV; p++) {
        i = permV[p];
        if (tid == 0) {
            sig[i] = 1;
            d[i] = 0;
            S[0] = i;
            start[0] = 0;
            end[0] = 1;
        }
        
        count = 1;
        numPhases = 0;
       
        #pragma omp barrier

        while (end[numPhases] - start[numPhases] > 0) {

            myCount = 0;
            #pragma omp barrier
            
            #pragma omp for schedule(dynamic)
            for (vert = start[numPhases]; vert < end[numPhases]; vert++) {
                
                v = S[vert];
                for (j=G->numEdges[v]; j<G->numEdges[v+1]; j++) {

                    /* Filter edges with weights divisible by 8 */
                    if ((G->weight[j] & 7) != 0) {
                        w = G->endV[j];
                        if (v != w) {
                            
                            myLock = omp_test_lock(&vLock[w]);
                           
                            if (myLock) { 
                                /* w found for the first time? */ 
                                if (d[w] == -1) {
                                    myS[myCount++] = w;
                                    d[w] = d[v] + 1;
                                    sig[w] = 0;
                                    sig[w] += sig[v];
                                    dynArrayInsert(&P[w], v);
                                } else if (d[w] == d[v] + 1) {
                                    sig[w] += sig[v];
                                    dynArrayInsert(&P[w], v);
                                }
                                
                                omp_unset_lock(&vLock[w]);
                            } else {
                                if ((d[w] == -1) || (d[w] == d[v]+ 1)) {
                                    omp_set_lock(&vLock[w]);
                                    sig[w] += sig[v];
                                    dynArrayInsert(&P[w], v);
                                    omp_unset_lock(&vLock[w]);
                                }
                            }
                           
                        }
                    }
                }
            }
            
            /* Merge all local stacks for next iteration */    
            numPhases++; 
            
            psCount[tid+1] = myCount;

            #pragma omp barrier
            
            if (tid == 0) {
                start[numPhases] = end[numPhases-1];
                psCount[0] = start[numPhases];
                for(k=1; k<=nthreads; k++) {
                    psCount[k] = psCount[k-1] + psCount[k];
                }
                end[numPhases] = psCount[nthreads];
            }
            #pragma omp barrier
            
            for (k = psCount[tid]; k < psCount[tid+1]; k++) {
                S[k] = myS[k-psCount[tid]];
            } 
            
            #pragma omp barrier
            count = end[numPhases];
        }
     
        numPhases--;
        
        #pragma omp barrier

        while (numPhases > 0) {
            #pragma omp for schedule(dynamic)
            for (j=start[numPhases]; j<end[numPhases]; j++) {
                w = S[j];
                for (k = 0; k<P[w].count; k++) {
                    v = P[w].vals[k];
                    omp_set_lock(&vLock[v]);
                    del[v] = del[v] + sig[v]*(1+del[w])/sig[w];
                    omp_unset_lock(&vLock[v]);
                }
                BC[w] += del[w];
            }
       
            #pragma omp barrier
            
            numPhases--;

            #pragma omp barrier
            
        }

        if (tid == 0) {
            chunkSize = count/nthreads;
        }
        
        #pragma omp barrier
       
        #pragma omp for schedule(static, chunkSize)
        for (j=0; j<count; j++) {
            w = S[j];
            d[w] = -1;
            del[w] = 0;
            P[w].count = 0;
        }

        #pragma omp barrier

    }

    #pragma omp for schedule(static, chunkSize)
    for (i=0; i<n; i++) {
        omp_destroy_lock(&vLock[i]);
    }
   
    free(myS);
    
    if (tid == 0) { 
        free(S);
        for (j=0; j<n; j++)
            freeDynArray(&P[j]);
        free(P);
        free(sig);
        free(d);
        free(del);
        free(vLock);
        free(start);
        free(end);
        free(psCount);
        time = omp_get_wtime() - time;
        free(permV);
    }

}    

    /* Verification */
    sum = 0;
    n = G->n;
    numV = 1<<K4Approx;
    for (i=0; i<n; i++) {
        sum += BC[i];
        // fprintf(stderr, "%lf ", BC[i]);
    }
    fprintf(stderr, "BC sum: %lf\n", sum); 
    
    return time;
}

